from enum import Enum
from django.conf import settings
from rest_framework import serializers
from rest_framework.exceptions import ValidationError
from arkindex_common.ml_tool import MLToolType
from arkindex_common.enums import TranscriptionType
from arkindex.project.serializer_fields import EnumField, DataSourceSlugField, PolygonField
from arkindex.documents.models import (
    Corpus, Element, ElementType, Transcription, DataSource, MLClass, Classification, ClassificationState
)
from arkindex.images.serializers import ZoneSerializer


class ClassificationMode(Enum):
    """
    Classification mode
    """
    Validate = "validate"
    Create = "create"


class DataSourceSerializer(serializers.ModelSerializer):
    """
    Serialize a data source for transcriptions and classifications
    """

    type = EnumField(MLToolType)

    class Meta:
        model = DataSource
        fields = (
            'id',
            'type',
            'slug',
            'name',
            'revision',
            'internal',
        )


class DataSourceStatsSerializer(DataSourceSerializer):
    """
    A data source, but including their result counts
    """
    classifications_count = serializers.IntegerField(default=0)
    transcriptions_count = serializers.IntegerField(default=0)
    entities_count = serializers.IntegerField(default=0)

    class Meta(DataSourceSerializer.Meta):
        fields = DataSourceSerializer.Meta.fields + (
            'classifications_count',
            'transcriptions_count',
            'entities_count',
        )


class MLClassSerializer(serializers.ModelSerializer):
    """
    Serializer for MLClass instances
    Used as a nested endpoint below corpus, so exclude corpus from payload
    """

    class Meta:
        model = MLClass
        exclude = ('corpus',)


class CountMLClassSerializer(MLClassSerializer):
    nb_best = serializers.IntegerField()


class ClassificationSerializer(serializers.ModelSerializer):
    """
    Serialize a classification on an Element
    """

    source = DataSourceSerializer()
    ml_class = MLClassSerializer()
    state = EnumField(ClassificationState)

    class Meta:
        model = Classification
        read_only_fields = ('id', 'confidence', 'high_confidence')
        fields = (
            'id',
            'source',
            'ml_class',
            'state',
            'confidence',
            'high_confidence'
        )


class ClassificationCreateSerializer(serializers.ModelSerializer):
    """
    Specific serializer for the manual classification creation
    """
    class Meta:
        model = Classification
        fields = (
            'element',
            'ml_class',
        )

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        if not self.context.get('request'):  # May be None when generating an OpenAPI schema or using from a REPL
            return

        # We need to use the default database to avoid stale read
        # when creating a classification on a newly created element
        self.fields['element'].queryset = Element.objects.using('default').filter(
            corpus__in=Corpus.objects.writable(self.context['request'].user))


class ClassificationsSelectionSerializer(serializers.ModelSerializer):
    """
    Serializer for the manual classification creation on the selection
    """
    ml_class = serializers.UUIDField(required=False)
    corpus_id = serializers.UUIDField()
    mode = EnumField(ClassificationMode)

    class Meta:
        model = Classification
        fields = (
            'corpus_id',
            'mode',
            'ml_class'
        )

    def validate(self, data):
        if not settings.ARKINDEX_FEATURES['selection']:
            raise ValidationError(['Selection is not available on this instance.'])
        data = super().validate(data)
        corpus = Corpus.objects.get(id=data['corpus_id'])

        if data['mode'] == ClassificationMode.Create:
            if 'ml_class' not in data:
                raise ValidationError({'ml_class': ['ML class is required when adding a classification']})
            ml_class = MLClass.objects.filter(id=data['ml_class']).first()
            if ml_class is None:
                raise ValidationError({'ml_class': 'Ml class {} not found'.format(data['ml_class'])})
            if ml_class.corpus.id != corpus.id:
                raise ValidationError('Ml class {} does not belong to the corpus {}'.format(
                    ml_class,
                    corpus
                ))
        return data


class TranscriptionSerializer(serializers.ModelSerializer):
    """
    Serialises a Transcription
    """
    type = EnumField(TranscriptionType, read_only=True)
    zone = ZoneSerializer(read_only=True)
    source = DataSourceSerializer(read_only=True)

    class Meta:
        model = Transcription
        read_only_fields = ('id', 'type', 'score', 'zone', 'source')
        fields = (
            'id',
            'type',
            'text',
            'score',
            'zone',
            'source',
        )


class TranscriptionCreateSerializer(serializers.ModelSerializer):
    """
    Allows the insertion of a manual transcription attached to an element
    """
    type = EnumField(TranscriptionType)
    source = DataSourceSlugField(tool_type=MLToolType.Recognizer)
    score = serializers.FloatField(min_value=0, max_value=1, required=False)

    class Meta:
        model = Transcription
        fields = ('text', 'type', 'source', 'score')

    def validate(self, data):
        data = super().validate(data)
        element = self.context.get('element')
        # Assert the transcribed element has a zone
        if not element.zone:
            raise ValidationError({'element': ['The element has no zone']})

        slug = data.get('source').slug
        if slug == 'manual':
            # Assert the type is allowed for manual transcription
            allowed_transcription = element.type.allowed_transcription
            if not allowed_transcription:
                raise ValidationError({'element': ['The element type does not allow creating a manual transcription']})

            if data['type'] is not allowed_transcription:
                raise ValidationError({'type': [
                    f"Only transcriptions of type '{allowed_transcription.value}' are allowed for this element"
                ]})
            return data

        # Additional validation for transcriptions with an internal source
        if not data.get('score'):
            raise ValidationError({'score': ['This field is required for non manual sources.']})

        user = self.context['request'].user
        if (not user or not user.is_internal):
            raise ValidationError({'source': [
                'An internal user is required to create a transcription with '
                f'the internal source "{slug}"'
            ]})
        return data


class TranscriptionBulkSerializer(serializers.Serializer):
    """
    Allows for insertion of new transcriptions and zones
    in Bulk (used by serializer below)
    Note: no element !
    """
    polygon = PolygonField()
    text = serializers.CharField()
    score = serializers.FloatField(min_value=0, max_value=1)
    type = EnumField(TranscriptionType)


class TranscriptionsSerializer(serializers.Serializer):
    """
    Allows for insertion of new transcriptions and zones
    in Bulk (uses serializer above) on a common parent
    """
    parent = serializers.PrimaryKeyRelatedField(
        # The real queryset is set in __init__
        queryset=Element.objects.none()
    )
    recognizer = DataSourceSlugField(tool_type=MLToolType.Recognizer)
    transcriptions = TranscriptionBulkSerializer(many=True, allow_empty=False)

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        if not self.context.get('request'):  # May be None when generating an OpenAPI schema or using from a REPL
            return

        # We need to use the default database to avoid stale read
        # when creating a transcription on a newly created element
        self.fields['parent'].queryset = Element.objects.using('default').filter(
            zone__isnull=False,
            corpus__in=Corpus.objects.writable(self.context['request'].user),
        )


class SimpleTranscriptionSerializer(serializers.Serializer):
    """
    Represent a unitary element transcription with no individual type
    Used to serialize transcriptions on elements in Bulk (used by serializer below)
    """
    polygon = PolygonField()
    text = serializers.CharField()
    score = serializers.FloatField(min_value=0, max_value=1)


class ElementTranscriptionsBulkSerializer(serializers.Serializer):
    """
    Serialize elements with their transcriptions on a common parent
    """
    element_type = serializers.SlugRelatedField(
        queryset=ElementType.objects.none(),
        slug_field='slug',
        help_text='An ElementType slug used to create all sub elements. Only one type is allowed per request'
    )
    transcription_type = EnumField(
        TranscriptionType,
        help_text='A TranscriptionType for created transcriptions'
    )
    source = DataSourceSlugField(
        tool_type=MLToolType.Recognizer,
        help_text='A recognizer DataSource slug, unique per request too. It cannot be set to manual'
    )
    transcriptions = SimpleTranscriptionSerializer(
        many=True,
        allow_empty=False,
        help_text='A list of transcriptions with their text, score and element zone'
    )

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        if not self.context.get('element'):  # May be None when generating an OpenAPI schema or using from a REPL
            return
        # Use the parent types for validation as elements are in the same corpus
        self.fields['element_type'].queryset = ElementType.objects.filter(corpus=self.context['element'].corpus)

    def validate_source(self, source):
        if source.slug == 'manual':
            raise ValidationError({'source': ["Transcriptions source slug cannot be set to manual."]})
        return source


class ClassificationBulkSerializer(serializers.Serializer):
    """
    Single classification serializer for bulk insertion
    Cannot use ModelSerializer as they become read-only when nested
    """
    class_name = serializers.CharField(source='ml_class')
    confidence = serializers.FloatField(min_value=0, max_value=1)
    high_confidence = serializers.BooleanField(default=False)


class ClassificationsSerializer(serializers.Serializer):
    """
    Insert N classifications on a single element from a single ML tool
    """
    parent = serializers.PrimaryKeyRelatedField(
        # The real queryset is set in __init__
        queryset=Element.objects.none(),
    )
    classifier = DataSourceSlugField(tool_type=MLToolType.Classifier)
    classifications = ClassificationBulkSerializer(many=True, allow_empty=False)

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        if not self.context.get('request'):  # May be None when generating an OpenAPI schema or using from a REPL
            return

        # We need to use the default database to avoid stale read
        # when creating a classification on a newly created element
        self.fields['parent'].queryset = Element.objects.using('default').filter(
            corpus__in=Corpus.objects.writable(self.context['request'].user),
        )

    def validate(self, data):
        data = super().validate(data)
        corpus_id = data['parent'].corpus.id
        for classification in data['classifications']:
            # Retrieve or create required ml_class
            classification['ml_class'], _ = MLClass.objects.get_or_create(
                name=classification['ml_class'],
                corpus_id=corpus_id
            )
        return data
