from django.urls import reverse
from rest_framework import status
from arkindex.project.tests import FixtureAPITestCase
from arkindex.project.tools import build_tree
from arkindex.documents.models import Element


class TestElementNeighbors(FixtureAPITestCase):

    @classmethod
    def setUpTestData(cls):
        super().setUpTestData()
        cls.volume_type = cls.corpus.types.get(slug='volume')
        cls.page_type = cls.corpus.types.get(slug='page')

    def test_element_neighbors(self):
        elements = build_tree(
            {
                'Y': 'Z',
                'B': 'X',
                'A': ['X', 'Y'],
                'C': 'X',
            },
            corpus=self.corpus,
            type=self.volume_type,
        )
        # Neighbors of A with a different type
        element = Element.objects.create(
            corpus=self.corpus,
            type=self.page_type,
            name='D',
        )
        element.add_parent(elements['Y'])
        element = Element.objects.create(
            corpus=self.corpus,
            type=self.page_type,
            name='E',
        )
        element.add_parent(elements['Y'])

        # Neighbors of A with the same type
        element = Element.objects.create(
            corpus=self.corpus,
            type=self.volume_type,
            name='F',
        )
        element.add_parent(elements['Y'])

        response = self.client.get(reverse('api:elements-neighbors', kwargs={'pk': str(elements['A'].id)}))
        self.assertEqual(response.status_code, status.HTTP_200_OK)
        results = response.json()['results']
        self.assertListEqual(
            results,
            sorted([{
                'position': 0,
                'element': {
                    'id': str(elements['B'].id),
                    'type': 'volume',
                    'name': elements['B'].name
                },
                'parents': [{
                    'id': str(elements['X'].id),
                    'type': 'volume',
                    'name': elements['X'].name
                }]
            }, {
                'position': 1,
                'element': {
                    'id': str(elements['A'].id),
                    'type': 'volume',
                    'name': elements['A'].name
                },
                'parents': [{
                    'id': str(elements['X'].id),
                    'type': 'volume',
                    'name': elements['X'].name
                }]
            }, {
                'position': 2,
                'element': {
                    'id': str(elements['C'].id),
                    'type': 'volume',
                    'name': elements['C'].name
                },
                'parents': [{
                    'id': str(elements['X'].id),
                    'type': 'volume',
                    'name': elements['X'].name
                }]
            }, {
                'position': 0,
                'element': {
                    'id': str(elements['A'].id),
                    'type': 'volume',
                    'name': elements['A'].name
                },
                'parents': [{
                    'id': str(elements['Z'].id),
                    'type': 'volume',
                    'name': elements['Z'].name
                }, {
                    'id': str(elements['Y'].id),
                    'type': 'volume',
                    'name': elements['Y'].name
                }]
            }, {
                'position': 1,
                'element': {
                    'id': str(element.id),
                    'type': 'volume',
                    'name': element.name
                },
                'parents': [{
                    'id': str(elements['Z'].id),
                    'type': 'volume',
                    'name': elements['Z'].name
                }, {
                    'id': str(elements['Y'].id),
                    'type': 'volume',
                    'name': elements['Y'].name
                }]
            }],
                # Order result by paths, as we can't predict the test generated uuids
                key=lambda path: ''.join([p['id'] for p in path['parents']]))
        )

    def test_n_element_neighbors(self):
        elements = build_tree(
            {
                'A': 'W',
                'B': 'W',
                'C': 'W',
            },
            corpus=self.corpus,
            type=self.volume_type,
        )

        response = self.client.get(
            reverse('api:elements-neighbors', kwargs={'pk': str(elements['A'].id)}),
            data={'n': 2})
        self.assertEqual(response.status_code, status.HTTP_200_OK)
        results = response.json()['results']
        self.assertListEqual(
            results,
            [{
                'position': 0,
                'element': {
                    'id': str(elements['A'].id),
                    'type': 'volume',
                    'name': elements['A'].name
                },
                'parents': [{
                    'id': str(elements['W'].id),
                    'type': 'volume',
                    'name': elements['W'].name
                }]
            }, {
                'position': 1,
                'element': {
                    'id': str(elements['B'].id),
                    'type': 'volume',
                    'name': 'B'
                },
                'parents': [{
                    'id': str(elements['W'].id),
                    'type': 'volume',
                    'name': elements['W'].name
                }]
            }, {
                'position': 2,
                'element': {
                    'id': str(elements['C'].id),
                    'type': 'volume',
                    'name': elements['C'].name
                },
                'parents': [{
                    'id': str(elements['W'].id),
                    'type': 'volume',
                    'name': elements['W'].name
                }]
            }]
        )

    def test_element_neighbors_missing_ordering(self):
        """
        Check ElementNeighbors handles non-sequential orderings
        """
        elements = build_tree(
            {
                'A': 'W',
                'B': 'W',
                'C': 'W',
            },
            corpus=self.corpus,
            type=self.volume_type,
        )
        c_path = elements['C'].paths.get()
        c_path.ordering = 6584
        c_path.save()

        response = self.client.get(reverse('api:elements-neighbors', kwargs={'pk': str(elements['B'].id)}))
        self.assertEqual(response.status_code, status.HTTP_200_OK)
        results = response.json()['results']
        self.assertListEqual(
            results,
            [{
                'position': 0,
                'element': {
                    'id': str(elements['A'].id),
                    'type': 'volume',
                    'name': elements['A'].name
                },
                'parents': [{
                    'id': str(elements['W'].id),
                    'type': 'volume',
                    'name': elements['W'].name
                }]
            }, {
                'position': 1,
                'element': {
                    'id': str(elements['B'].id),
                    'type': 'volume',
                    'name': 'B'
                },
                'parents': [{
                    'id': str(elements['W'].id),
                    'type': 'volume',
                    'name': elements['W'].name
                }]
            }, {
                'position': 6584,
                'element': {
                    'id': str(elements['C'].id),
                    'type': 'volume',
                    'name': elements['C'].name
                },
                'parents': [{
                    'id': str(elements['W'].id),
                    'type': 'volume',
                    'name': elements['W'].name
                }]
            }]
        )

    def test_neighbors_n_limit(self):
        bad_values = ['lol', 'nan', 'inf', '-inf', 42, 0, -7]
        for value in bad_values:
            response = self.client.get(
                reverse('api:elements-neighbors', kwargs={'pk': str(self.corpus.elements.first().id)}),
                data={'n': value},
            )
            self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)

    def test_sub_neighbors(self):
        r"""
        Ensure the children of some elements of the same type do not interfere with the neighboring elements
           V
          /|\
         A B C
          / \
         B1 B2
        """
        elements = build_tree(
            {
                'A': 'V',
                'B': 'V',
                'C': 'V',
                'B1': 'B',
                'B2': 'B',
            },
            corpus=self.corpus,
            type=self.volume_type,
        )
        response = self.client.get(reverse('api:elements-neighbors', kwargs={'pk': str(elements['B'].id)}))
        self.assertEqual(response.status_code, status.HTTP_200_OK)
        results = response.json()['results']

        self.assertListEqual(
            results,
            [{
                'position': 0,
                'element': {
                    'id': str(elements['A'].id),
                    'type': 'volume',
                    'name': elements['A'].name
                },
                'parents': [{
                    'id': str(elements['V'].id),
                    'type': 'volume',
                    'name': elements['V'].name
                }]
            }, {
                'position': 1,
                'element': {
                    'id': str(elements['B'].id),
                    'type': 'volume',
                    'name': 'B'
                },
                'parents': [{
                    'id': str(elements['V'].id),
                    'type': 'volume',
                    'name': elements['V'].name
                }]
            }, {
                'position': 2,
                'element': {
                    'id': str(elements['C'].id),
                    'type': 'volume',
                    'name': elements['C'].name
                },
                'parents': [{
                    'id': str(elements['V'].id),
                    'type': 'volume',
                    'name': elements['V'].name
                }]
            }]
        )

    def test_not_found(self):
        self.assertFalse(Element.objects.filter(id='aaaaaaaa-aaaa-aaaa-aaaa-aaaaaaaaaaaa').exists())
        response = self.client.get(reverse(
            'api:elements-neighbors',
            kwargs={'pk': 'aaaaaaaa-aaaa-aaaa-aaaa-aaaaaaaaaaaa'},
        ))
        self.assertEqual(response.status_code, status.HTTP_404_NOT_FOUND)
