from django.urls import reverse
from django.test import override_settings
from unittest.mock import patch
from mock import AsyncMock
from rest_framework import status
from arkindex.project.tests import FixtureAPITestCase
from arkindex.project.polygon import Polygon
from arkindex_common.enums import TranscriptionType
from arkindex_common.ml_tool import MLToolType
from arkindex.documents.models import Element, Corpus, Transcription, DataSource
from arkindex.users.models import User
from uuid import uuid4


class TestTranscriptionCreate(FixtureAPITestCase):
    """
    Tests for text element creation view
    """

    @classmethod
    def setUpTestData(cls):
        super().setUpTestData()
        cls.page = cls.corpus.elements.get(name='Volume 1, page 1r')
        cls.vol = cls.corpus.elements.get(name='Volume 1')
        cls.src = DataSource.objects.get(slug='test')
        cls.line = cls.corpus.elements.filter(type__slug='text_line').first()
        cls.ml_transcription = cls.page.transcriptions.get(text='PARIS')
        cls.private_corpus = Corpus.objects.create(name='Private')
        cls.private_page = cls.private_corpus.elements.create(type=cls.page.type)
        # Create an user with a read right only on the private corpus
        cls.private_read_user = User.objects.create_user('a@bc.de', 'a')
        cls.private_read_user.verified_email = True
        cls.private_read_user.save()
        cls.private_corpus.corpus_right.create(user=cls.private_read_user)
        # Create an internal user
        cls.internal_user = User.objects.create_internal_user('internal@internal.fr')
        cls.internal_user.verified_email = True
        cls.internal_user.save()

    def setUp(self):
        self.manual_source = DataSource.objects.create(type=MLToolType.Recognizer, slug='manual', internal=False)
        self.manual_transcription = self.line.transcriptions.create(
            text='A manual transcription',
            source=self.manual_source,
            type=TranscriptionType.Line
        )
        self.private_transcription = self.private_page.transcriptions.create(
            text='PEPE',
            type=TranscriptionType.Line,
            source=self.manual_source
        )

    def test_create_transcription_require_login(self):
        response = self.client.post(reverse('api:transcription-create', kwargs={'pk': self.line.id}), format='json')
        self.assertEqual(response.status_code, status.HTTP_403_FORBIDDEN)
        self.assertDictEqual(response.json(), {
            'detail': 'Authentication credentials were not provided.'
        })

    def test_create_transcription_write_right(self):
        self.client.force_login(self.private_read_user)
        response = self.client.post(
            reverse('api:transcription-create', kwargs={'pk': self.private_page.id}),
            format='json',
            data={
                'type': 'word',
                'text': 'NEKUDOTAYIM',
                'source': 'manual'
            }
        )
        self.assertEqual(response.status_code, status.HTTP_404_NOT_FOUND)

    @patch('arkindex.project.triggers.get_channel_layer')
    def test_create_transcription_no_element(self, get_layer_mock):
        self.client.force_login(self.user)
        response = self.client.post(
            reverse('api:transcription-create', kwargs={'pk': uuid4()}),
            format='json',
            data={
                'type': 'word',
                'text': 'NEKUDOTAYIM',
                'source': 'manual'
            }
        )
        self.assertEqual(response.status_code, status.HTTP_404_NOT_FOUND)

    @patch('arkindex.project.triggers.get_channel_layer')
    def test_create_transcription_elt_null_zone(self, get_layer_mock):
        null_zone_page = self.corpus.elements.create(type=self.page.type)

        self.client.force_login(self.user)
        response = self.client.post(
            reverse('api:transcription-create', kwargs={'pk': null_zone_page.id}),
            format='json',
            data={
                'type': 'word',
                'text': 'NEKUDOTAYIM',
                'source': 'manual'
            }
        )
        self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)
        self.assertDictEqual(response.json(), {
            'element': ['The element has no zone']
        })

    @patch('arkindex.project.triggers.get_channel_layer')
    def test_create_transcription(self, get_layer_mock):
        """
        Checks the view creates a manual transcription and runs ES indexing
        """
        get_layer_mock.return_value.send = AsyncMock()

        # The view must have the ability to create a manual source if it does not exist
        self.manual_source.delete()
        self.assertFalse(DataSource.objects.filter(type=MLToolType.Recognizer, slug='manual').exists())

        self.client.force_login(self.user)
        with self.assertNumQueries(11):
            response = self.client.post(
                reverse('api:transcription-create', kwargs={'pk': self.line.id}),
                format='json',
                data={
                    'type': 'line',
                    'text': 'A perfect day in a perfect place',
                    'source': 'manual'
                }
            )
        self.assertEqual(response.status_code, status.HTTP_201_CREATED)
        tr = Transcription.objects.get(text='A perfect day in a perfect place')
        self.assertDictEqual(response.json(), {
            'id': str(tr.id),
            'score': None,
            'source': {
                'id': str(tr.source.id),
                'internal': False,
                'name': '',
                'revision': '',
                'slug': 'manual',
                'type': 'recognizer'
            },
            'text': 'A perfect day in a perfect place',
            'type': 'line',
            'zone': None
        })

        new_ts = Transcription.objects.get(text='A perfect day in a perfect place', type=TranscriptionType.Line)
        self.assertIsNone(new_ts.zone)
        self.assertIsNone(new_ts.score)
        self.assertTrue(DataSource.objects.filter(type=MLToolType.Recognizer, slug='manual').exists())
        self.assertEqual(new_ts.source.slug, 'manual')
        self.assertTrue(self.line.transcriptions.filter(pk=new_ts.id).exists())

        get_layer_mock().send.assert_called_once_with('reindex', {
            'type': 'reindex.start',
            'element': str(self.line.id),
            'corpus': None,
            'entity': None,
            'transcriptions': True,
            'elements': True,
            'entities': False,
            'drop': False,
        })

    @patch('arkindex.project.triggers.get_channel_layer')
    def test_create_transcription_takes_no_zone(self, get_layer_mock):
        """
        TranscriptionCreate endpoint do not create any zone for the transcription
        """
        get_layer_mock.return_value.send = AsyncMock()

        self.client.force_login(self.user)
        response = self.client.post(
            reverse('api:transcription-create', kwargs={'pk': self.line.id}),
            format='json',
            data={
                'type': 'line',
                'polygon': [(0, 0), (42, 0), (42, 42), (0, 42), (0, 0)],
                'text': 'SQUARE',
                'source': 'manual'
            }
        )
        self.assertEqual(response.status_code, status.HTTP_201_CREATED)

        new_ts = Transcription.objects.get(text='SQUARE', type=TranscriptionType.Line)
        self.assertEqual(new_ts.zone, None)

    @patch('arkindex.project.triggers.get_channel_layer')
    def test_create_duplicated_transcription(self, get_layer_mock):
        """
        Check the view creates a new transcriptions with a similar manual source
        """
        get_layer_mock.return_value.send = AsyncMock()

        self.client.force_login(self.user)
        ts = self.page.transcriptions.create(
            text='GLOUBIBOULGA',
            type=TranscriptionType.Word,
            source=self.manual_source
        )
        with self.assertNumQueries(8):
            response = self.client.post(
                reverse('api:transcription-create', kwargs={'pk': self.line.id}),
                format='json',
                data={
                    'type': 'line',
                    'text': ts.text,
                    'source': 'manual'
                }
            )
        self.assertEqual(response.status_code, status.HTTP_201_CREATED)
        self.assertEqual(Transcription.objects.filter(text=ts.text, source=self.manual_source).count(), 2)
        ts.refresh_from_db()
        self.assertNotEqual(ts.score, 0.99)

    @patch('arkindex.project.triggers.get_channel_layer')
    def test_create_transcription_wrong_type(self, get_layer_mock):
        self.client.force_login(self.user)
        response = self.client.post(
            reverse('api:transcription-create', kwargs={'pk': self.line.id}),
            format='json',
            data={
                'type': 'AAAAA',
                'text': 'NEKUDOTAYIM',
                'source': 'manual'
            }
        )
        self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)
        self.assertDictEqual(response.json(), {
            'type': ['Value is not of type TranscriptionType']
        })

    @patch('arkindex.project.triggers.get_channel_layer')
    @override_settings(ARKINDEX_FEATURES={'search': False})
    def test_create_transcription_no_search(self, get_layer_mock):
        self.client.force_login(self.user)
        response = self.client.post(
            reverse('api:transcription-create', kwargs={'pk': self.line.id}),
            format='json',
            data={
                'type': 'line',
                'text': 'A classy text line',
                'source': 'manual'
            }
        )
        self.assertEqual(response.status_code, status.HTTP_201_CREATED)
        self.assertFalse(get_layer_mock().send.called)

    def test_manual_transcription_forbidden_type(self):
        """
        Creating a manual transcription with a non allowed type is forbidden
        """
        self.client.force_login(self.user)
        response = self.client.post(
            reverse('api:transcription-create', kwargs={'pk': self.line.id}),
            format='json',
            data={
                'type': 'word',
                'text': 'NEKUDOTAYIM',
                'source': 'manual'
            }
        )
        self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)
        self.assertDictEqual(response.json(), {
            'type': ["Only transcriptions of type 'line' are allowed for this element"]
        })

    def test_manual_transcription_no_allowed_type(self):
        self.line.type.allowed_transcription = None
        self.line.type.save()
        self.client.force_login(self.user)
        response = self.client.post(
            reverse('api:transcription-create', kwargs={'pk': self.line.id}),
            format='json',
            data={
                'type': 'line',
                'text': 'A classy text line',
                'source': 'manual'
            }
        )
        self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)
        self.assertDictEqual(response.json(), {
            'element': ['The element type does not allow creating a manual transcription']
        })

    def test_transcription_retrieve(self):
        """
        Allows to retrieve a single transcription with a read right
        """
        self.client.force_login(self.private_read_user)
        response = self.client.get(reverse('api:transcription-edit', kwargs={'pk': self.manual_transcription.id}))
        self.assertEqual(response.status_code, status.HTTP_200_OK)
        tr = Transcription.objects.get(id=response.json()['id'])
        self.manual_source.refresh_from_db()
        self.assertDictEqual(response.json(), {
            'id': str(tr.id),
            'score': None,
            'source': {
                'id': str(self.manual_source.id),
                'internal': False,
                'name': '',
                'revision': '',
                'slug': 'manual',
                'type': 'recognizer'
            },
            'text': 'A manual transcription',
            'type': 'line',
            'zone': None
        })

    def test_transcription_retrieve_ml(self):
        """
        Retrieve a transcription produced by a ML worker
        """
        self.client.force_login(self.user)
        response = self.client.get(reverse('api:transcription-edit', kwargs={'pk': self.ml_transcription.id}))
        self.assertEqual(response.status_code, status.HTTP_200_OK)

    def test_transcription_retrieve_verified_user(self):
        response = self.client.get(reverse('api:transcription-edit', kwargs={'pk': self.manual_transcription.id}))
        self.assertEqual(response.status_code, status.HTTP_403_FORBIDDEN)
        self.assertDictEqual(response.json(), {'detail': 'Authentication credentials were not provided.'})

    def test_transcription_retrieve_private(self):
        user = User.objects.create_user('f@gh.ij', 'b')
        user.verified_email = True
        user.save()
        self.client.force_login(user)
        response = self.client.get(reverse('api:transcription-edit', kwargs={'pk': self.private_transcription.id}))
        self.assertEqual(response.status_code, status.HTTP_404_NOT_FOUND)

    def test_transcription_patch(self):
        self.client.force_login(self.user)
        response = self.client.patch(
            reverse('api:transcription-edit', kwargs={'pk': self.manual_transcription.id}),
            format='json',
            data={'text': 'Once upon a time in a castle'}
        )
        self.assertEqual(response.status_code, status.HTTP_200_OK)
        data = response.json()
        self.assertEqual(data.get('id'), str(self.manual_transcription.id))
        self.assertEqual(data.get('text'), 'Once upon a time in a castle')

    def test_transcription_patch_text_only(self):
        """
        Assert only transcription text can be updated
        """
        self.client.force_login(self.user)
        manual_tr_id = self.manual_transcription.id
        response = self.client.patch(
            reverse('api:transcription-edit', kwargs={'pk': manual_tr_id}),
            format='json',
            data={
                'id': uuid4(),
                'text': 'a knight was living lonely',
                'type': 'page',
                'score': .42,
                'zone': {'polygon': [[4, 2], [2, 4], [3, 3]], 'center': [42, 1337]},
                'image': {'status': 'checked'},
                'source': 'castle_source',
                'element': 'water'
            }
        )
        self.assertEqual(response.status_code, status.HTTP_200_OK)
        self.manual_transcription.refresh_from_db()
        self.assertDictEqual(response.json(), {
            'id': str(manual_tr_id),
            'score': None,
            'source': {
                'id': str(self.manual_source.id),
                'internal': False,
                'name': '',
                'revision': '',
                'slug': 'manual',
                'type': 'recognizer'
            },
            'text': 'a knight was living lonely',
            'type': 'line',
            'zone': None
        })

    def test_transcription_patch_write_right(self):
        """
        A write right is required to patch a manual transcription
        """
        self.client.force_login(self.private_read_user)
        response = self.client.patch(
            reverse('api:transcription-edit', kwargs={'pk': self.private_transcription.id}),
            format='json',
            data={'text': 'He was good and fearless'}
        )
        self.assertEqual(response.status_code, status.HTTP_403_FORBIDDEN)
        self.assertDictEqual(response.json(), {
            '__all__': ['A write access to transcription element corpus is required.']
        })

    def test_transcription_patch_not_manual(self):
        """
        Updating a transcription produced by a ML worker is forbidden
        """
        self.client.force_login(self.user)
        response = self.client.patch(
            reverse('api:transcription-edit', kwargs={'pk': self.ml_transcription.id}),
            format='json',
            data={'text': 'He used to play the Bumbulum'}
        )
        self.assertEqual(response.status_code, status.HTTP_403_FORBIDDEN)
        self.assertDictEqual(response.json(), {
            '__all__': ['Only manual transcriptions can be edited.']
        })

    def test_manual_transcription_delete(self):
        """
        Allows to delete a manual transcription with a write right
        """
        self.client.force_login(self.user)
        response = self.client.delete(reverse('api:transcription-edit', kwargs={'pk': self.manual_transcription.id}))
        self.assertEqual(response.status_code, status.HTTP_204_NO_CONTENT)
        self.assertFalse(Transcription.objects.filter(id=self.manual_transcription.id).exists())

    def test_manual_transcription_delete_write_right(self):
        """
        Deleting a transcription without a write right is forbidden
        """
        self.client.force_login(self.private_read_user)
        response = self.client.delete(reverse('api:transcription-edit', kwargs={'pk': self.manual_transcription.id}))
        self.assertEqual(response.status_code, status.HTTP_403_FORBIDDEN)
        self.assertDictEqual(response.json(), {
            '__all__': ['A write access to transcription element corpus is required.']
        })

    @patch('arkindex.project.serializer_fields.MLTool.get')
    def test_ml_transcription_non_internal(self, ml_get_mock):
        """
        A transcription created on an element with an internal source
        should only be authorized for an internal user (e.g. a worker)
        """
        ml_get_mock.return_value.type = self.src.type
        ml_get_mock.return_value.slug = self.src.slug
        ml_get_mock.return_value.name = self.src.name
        ml_get_mock.return_value.version = self.src.revision
        self.src.internal = True
        self.src.save()

        self.client.force_login(self.user)
        response = self.client.post(
            reverse('api:transcription-create', kwargs={'pk': self.line.id}),
            format='json',
            data={
                'type': 'word',
                'text': 'SQUARE',
                'source': self.src.slug,
                'score': .99
            }
        )
        self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)
        self.assertEqual(response.json(), {'source': [
            'An internal user is required to create a transcription with the internal source "test"']
        })

    @patch('arkindex.project.serializer_fields.MLTool.get')
    def test_ml_trancription_required_score(self, ml_get_mock):
        """
        A score is required when creating a transcription on an
        element with an internal source
        """
        ml_get_mock.return_value.type = self.src.type
        ml_get_mock.return_value.slug = self.src.slug
        ml_get_mock.return_value.name = self.src.name
        ml_get_mock.return_value.version = self.src.revision
        self.src.internal = True
        self.src.save()

        self.client.force_login(self.internal_user)
        response = self.client.post(
            reverse('api:transcription-create', kwargs={'pk': self.line.id}),
            format='json',
            data={
                'type': 'word',
                'text': 'CIRCLE',
                'source': self.src.slug
            }
        )
        self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)
        self.assertEqual(response.json(), {'score': ['This field is required for non manual sources.']})

    @patch('arkindex.project.serializer_fields.MLTool.get')
    def test_ml_trancription_requires_element_zone(self, ml_get_mock):
        """
        A worker can publish a transcription on an element with a zone only
        """
        null_zone_page = self.corpus.elements.create(type=self.page.type)
        ml_get_mock.return_value.type = self.src.type
        ml_get_mock.return_value.slug = self.src.slug
        ml_get_mock.return_value.name = self.src.name
        ml_get_mock.return_value.version = self.src.revision
        self.src.internal = True
        self.src.save()

        self.client.force_login(self.internal_user)
        response = self.client.post(
            reverse('api:transcription-create', kwargs={'pk': null_zone_page.id}),
            format='json',
            data={
                'type': 'word',
                'text': 'ELLIPSE',
                'source': self.src.slug,
                'score': .42
            }
        )
        self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)
        self.assertEqual(response.json(), {'element': ['The element has no zone']})

    @patch('arkindex.project.serializer_fields.MLTool.get')
    @patch('arkindex.project.triggers.get_channel_layer')
    def test_ml_transcription_create(self, get_layer_mock, ml_get_mock):
        """
        Creates a transcription depending on an element zone
        """
        get_layer_mock.return_value.send = AsyncMock()
        ml_get_mock.return_value.type = self.src.type
        ml_get_mock.return_value.slug = self.src.slug
        ml_get_mock.return_value.name = self.src.name
        ml_get_mock.return_value.version = self.src.revision
        self.src.internal = True
        self.src.save()

        self.client.force_login(self.internal_user)
        self.assertTrue(self.internal_user.is_internal)
        response = self.client.post(
            reverse('api:transcription-create', kwargs={'pk': self.line.id}),
            format='json',
            data={
                'type': 'word',
                'text': 'TRIANGLE',
                'source': self.src.slug,
                'score': .42
            }
        )
        self.assertEqual(response.status_code, status.HTTP_201_CREATED)
        self.assertEqual(self.line.transcriptions.filter(source=self.src).count(), 1)

        get_layer_mock().send.assert_called_once_with('reindex', {
            'type': 'reindex.start',
            'element': str(self.line.id),
            'corpus': None,
            'entity': None,
            'transcriptions': True,
            'elements': True,
            'entities': False,
            'drop': False,
        })

    @patch('arkindex.project.serializer_fields.MLTool.get')
    @patch('arkindex.project.triggers.get_channel_layer')
    def test_create_bulk_transcription(self, get_layer_mock, ml_get_mock):
        """
        Checks the view creates transcriptions, zones, links, paths and runs ES indexing
        Using bulk_transcriptions
        """
        get_layer_mock.return_value.send = AsyncMock()
        ml_get_mock.return_value.type = self.src.type
        ml_get_mock.return_value.slug = self.src.slug
        ml_get_mock.return_value.name = self.src.name
        ml_get_mock.return_value.version = self.src.revision
        self.src.internal = True
        self.src.save()

        self.page.transcriptions.all().delete()
        self.client.force_login(self.internal_user)
        request_data = {
            "parent": str(self.page.id),
            "recognizer": self.src.slug,
            "transcriptions": [
                {
                    "type": "word",
                    "polygon": [(0, 0), (100, 0), (100, 100), (0, 100), (0, 0)],
                    "text": "NEKUDOTAYIM",
                    "score": 0.83,
                },
                {
                    "type": "line",
                    "polygon": [(0, 0), (200, 0), (200, 200), (0, 200), (0, 0)],
                    "text": "This is a test",
                    "score": 0.75,
                },
            ]
        }
        response = self.client.post(reverse('api:transcription-bulk'), format='json', data=request_data)
        self.assertEqual(response.status_code, status.HTTP_201_CREATED)
        self.assertEqual(self.page.transcriptions.count(), 2)

        word_ts = self.page.transcriptions.get(text="NEKUDOTAYIM", type=TranscriptionType.Word)
        self.assertEqual(word_ts.zone.polygon, Polygon.from_coords(0, 0, 100, 100))
        self.assertEqual(word_ts.score, 0.83)
        self.assertEqual(word_ts.source, self.src)

        line_ts = self.page.transcriptions.get(text="This is a test", type=TranscriptionType.Line)
        self.assertEqual(line_ts.zone.polygon, Polygon.from_coords(0, 0, 200, 200))
        self.assertEqual(line_ts.score, 0.75)
        self.assertEqual(line_ts.source, self.src)

        get_layer_mock().send.assert_called_once_with('reindex', {
            'type': 'reindex.start',
            'corpus': None,
            'element': str(self.page.id),
            'entity': None,
            'transcriptions': True,
            'elements': True,
            'entities': False,
            'drop': False,
        })

    @patch('arkindex.project.serializer_fields.MLTool.get')
    def test_bulk_transcription_no_zone(self, ml_get_mock):
        self.assertIsNone(self.vol.zone)

        ml_get_mock.return_value.type = self.src.type
        ml_get_mock.return_value.slug = self.src.slug
        ml_get_mock.return_value.name = self.src.name
        ml_get_mock.return_value.version = self.src.revision
        self.src.internal = True
        self.src.save()

        self.client.force_login(self.user)
        response = self.client.post(reverse('api:transcription-bulk'), format='json', data={
            "parent": str(self.vol.id),
            "recognizer": self.src.slug,
            "transcriptions": [
                {
                    "type": "word",
                    "polygon": [(0, 0), (100, 0), (100, 100), (0, 100), (0, 0)],
                    "text": "NEKUDOTAYIM",
                    "score": 0.83,
                },
            ]
        })
        self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)

    @patch('arkindex.project.serializer_fields.MLTool.get')
    @patch('arkindex.project.triggers.get_channel_layer')
    def test_update_bulk_transcription(self, get_layer_mock, ml_get_mock):
        get_layer_mock.return_value.send = AsyncMock()
        ml_get_mock.return_value.type = self.src.type
        ml_get_mock.return_value.slug = self.src.slug
        ml_get_mock.return_value.name = self.src.name
        ml_get_mock.return_value.version = self.src.revision
        self.src.internal = True
        self.src.save()

        data = {
            "parent": str(self.page.id),
            "recognizer": self.src.slug,
            "transcriptions": [
                {
                    "type": "word",
                    "polygon": [(0, 0), (100, 0), (100, 100), (0, 100), (0, 0)],
                    "text": "NEKUDOTAYIM",
                    "score": 0.83,
                },
                {
                    "type": "line",
                    "polygon": [(0, 0), (200, 0), (200, 200), (0, 200), (0, 0)],
                    "text": "This is a test",
                    "score": 0.75,
                },
            ]
        }

        self.client.force_login(self.user)
        response = self.client.put(reverse('api:transcription-bulk'), format='json', data=data)
        self.assertEqual(response.status_code, status.HTTP_201_CREATED)
        new_ts = self.page.transcriptions.get(text="NEKUDOTAYIM", type=TranscriptionType.Word)
        self.assertEqual(new_ts.zone.polygon, Polygon.from_coords(0, 0, 100, 100))
        self.assertEqual(new_ts.score, 0.83)
        self.assertEqual(new_ts.source, self.src)

        get_layer_mock().send.assert_called_once_with('reindex', {
            'type': 'reindex.start',
            'corpus': None,
            'element': str(self.page.id),
            'entity': None,
            'transcriptions': True,
            'elements': True,
            'entities': False,
            'drop': False,
        })

        # Update again
        get_layer_mock().send.reset_mock()
        data['transcriptions'] = [
            {
                "type": "word",
                "polygon": [(0, 0), (100, 0), (100, 100), (0, 100), (0, 0)],
                "text": "NEKUDOTAYIM",
                "score": 0.83,
            },
            {
                "type": "word",
                "polygon": [(0, 0), (200, 0), (200, 200), (0, 200), (0, 0)],
                "text": "Hi",
                "score": 0.99,
            },
        ]

        response = self.client.put(reverse('api:transcription-bulk'), format='json', data=data)
        self.assertEqual(response.status_code, status.HTTP_201_CREATED)
        get_layer_mock().send.assert_called_once_with('reindex', {
            'type': 'reindex.start',
            'corpus': None,
            'element': str(self.page.id),
            'entity': None,
            'transcriptions': True,
            'elements': True,
            'entities': False,
            'drop': False,
        })

        # Previous transcriptions should be replaced by two Word transcriptions
        self.assertEqual(self.page.transcriptions.count(), 2)

        # This transcription was inserted with the first call.
        # This ensures calling the update API with an existing transcription
        # keeps the existing transcriptions
        ts_1 = self.page.transcriptions.get(text="NEKUDOTAYIM", type=TranscriptionType.Word)
        self.assertEqual(ts_1.zone.polygon, Polygon.from_coords(0, 0, 100, 100))
        self.assertEqual(ts_1.score, 0.83)
        self.assertEqual(ts_1.source, self.src)

        ts_2 = self.page.transcriptions.get(text="Hi", type=TranscriptionType.Word)
        self.assertEqual(ts_2.zone.polygon, Polygon.from_coords(0, 0, 200, 200))
        self.assertEqual(ts_2.score, 0.99)
        self.assertEqual(ts_2.source, self.src)

    @patch('arkindex.project.serializer_fields.MLTool.get')
    @patch('arkindex.project.triggers.get_channel_layer')
    @override_settings(ARKINDEX_FEATURES={'search': False})
    def test_create_bulk_transcription_no_search(self, get_layer_mock, ml_get_mock):
        ml_get_mock.return_value.type = self.src.type
        ml_get_mock.return_value.slug = self.src.slug
        ml_get_mock.return_value.name = self.src.name
        ml_get_mock.return_value.version = self.src.revision
        self.src.internal = True
        self.src.save()

        self.client.force_login(self.user)
        request_data = {
            "parent": str(self.page.id),
            "recognizer": self.src.slug,
            "transcriptions": [
                {
                    "type": "word",
                    "polygon": [(0, 0), (100, 0), (100, 100), (0, 100), (0, 0)],
                    "text": "NEKUDOTAYIM",
                    "score": 0.83,
                },
                {
                    "type": "line",
                    "polygon": [(0, 0), (200, 0), (200, 200), (0, 200), (0, 0)],
                    "text": "This is a test",
                    "score": 0.75,
                },
            ]
        }
        response = self.client.post(reverse('api:transcription-bulk'), format='json', data=request_data)
        self.assertEqual(response.status_code, status.HTTP_201_CREATED)
        self.assertFalse(get_layer_mock().send.called)

    @patch('arkindex.project.serializer_fields.MLTool.get')
    @patch('arkindex.project.triggers.get_channel_layer')
    @override_settings(ARKINDEX_FEATURES={'search': False})
    def test_update_bulk_transcription_no_search(self, get_layer_mock, ml_get_mock):
        ml_get_mock.return_value.type = self.src.type
        ml_get_mock.return_value.slug = self.src.slug
        ml_get_mock.return_value.name = self.src.name
        ml_get_mock.return_value.version = self.src.revision
        self.src.internal = True
        self.src.save()

        data = {
            "parent": str(self.page.id),
            "recognizer": self.src.slug,
            "transcriptions": [
                {
                    "type": "word",
                    "polygon": [(0, 0), (100, 0), (100, 100), (0, 100), (0, 0)],
                    "text": "NEKUDOTAYIM",
                    "score": 0.83,
                },
                {
                    "type": "line",
                    "polygon": [(0, 0), (200, 0), (200, 200), (0, 200), (0, 0)],
                    "text": "This is a test",
                    "score": 0.75,
                },
            ]
        }

        self.client.force_login(self.user)
        response = self.client.put(reverse('api:transcription-bulk'), format='json', data=data)
        self.assertEqual(response.status_code, status.HTTP_201_CREATED)
        self.assertFalse(get_layer_mock().send.called)

    @patch('arkindex.project.serializer_fields.MLTool.get')
    @patch('arkindex.project.triggers.get_channel_layer')
    @override_settings(ARKINDEX_FEATURES={'search': False})
    def test_bulk_unique_zone(self, get_layer_mock, ml_get_mock):
        """
        Checks the bulk view reuses zones when available
        """
        ml_get_mock.return_value.type = self.src.type
        ml_get_mock.return_value.slug = 'new_slug'
        ml_get_mock.return_value.name = self.src.name
        ml_get_mock.return_value.version = self.src.revision
        self.src.internal = True
        self.src.save()

        self.assertEqual(self.page.transcriptions.count(), 4)
        existing_ts = self.page.transcriptions.get(zone__image__path='img1', text="PARIS")
        data = {
            "parent": str(self.page.id),
            "recognizer": self.src.slug,
            "transcriptions": [{
                "type": "word",
                "text": "PARISH",
                "score": 0.999,
                "polygon": existing_ts.zone.polygon.serialize(),
            }]
        }
        self.client.force_login(self.user)
        response = self.client.put(reverse('api:transcription-bulk'), format='json', data=data)
        self.assertEqual(response.status_code, status.HTTP_201_CREATED)

        # A new transcription has been added with the same zone
        self.assertEqual(self.page.transcriptions.count(), 5)
        self.assertEqual(Transcription.objects.get(text="PARISH").zone.id, existing_ts.zone.id)

    @patch('arkindex.project.serializer_fields.MLTool.get')
    @patch('arkindex.project.triggers.get_channel_layer')
    def test_bulk_transcriptions_v2(self, get_layer_mock, ml_get_mock):
        """
        Bulk creates a list of element with an attached transcription
        """
        get_layer_mock.return_value.send = AsyncMock()
        ml_get_mock.return_value.type = self.src.type
        ml_get_mock.return_value.slug = self.src.slug
        ml_get_mock.return_value.name = self.src.name
        ml_get_mock.return_value.version = self.src.revision
        self.src.internal = True
        self.src.save()

        transcriptions = [
            ([[13, 37], [133, 37], [133, 137], [13, 137], [13, 37]], 'Hello world !', 0.1337),
            ([[24, 42], [64, 42], [64, 142], [24, 142], [24, 42]], 'I <3 JavaScript', 0.42),
        ]
        data = {
            'element_type': 'text_line',
            'transcription_type': 'line',
            'source': self.src.slug,
            'transcriptions': [{
                'polygon': poly,
                'text': text,
                'score': score
            } for poly, text, score in transcriptions]
        }

        self.client.force_login(self.internal_user)
        with self.assertNumQueries(20):
            response = self.client.post(
                reverse('api:element-transcriptions-bulk', kwargs={'pk': self.page.id}),
                format='json',
                data=data
            )
        self.assertEqual(response.status_code, status.HTTP_201_CREATED)

        created_elts = Element.objects.get_descending(self.page.id).exclude(id=self.line.id)
        self.assertEqual(created_elts.count(), 2)
        self.assertTrue(all(map(lambda elt: elt.zone.image == self.page.zone.image, created_elts)))
        self.assertCountEqual(
            map(lambda elt: (elt.paths.first().ordering, elt.name, elt.zone.polygon), created_elts),
            [
                (1, '2', Polygon([[13, 37], [133, 37], [133, 137], [13, 137], [13, 37]])),
                (2, '3', Polygon([[24, 42], [64, 42], [64, 142], [24, 142], [24, 42]]))
            ]
        )
        self.assertCountEqual(
            created_elts.values_list('transcriptions__type', 'transcriptions__text', 'transcriptions__zone'),
            [
                (TranscriptionType.Line, ('Hello world !'), None),
                (TranscriptionType.Line, ('I <3 JavaScript'), None)
            ]
        )
        get_layer_mock().send.assert_called_once_with('reindex', {
            'type': 'reindex.start',
            'corpus': None,
            'element': str(self.page.id),
            'entity': None,
            'transcriptions': True,
            'elements': True,
            'entities': False,
            'drop': False,
        })

    @patch('arkindex.project.serializer_fields.MLTool.get')
    @override_settings(ARKINDEX_FEATURES={'search': False})
    def test_bulk_transcriptions_v2_load(self, ml_get_mock):
        """
        Sending 100 transcriptions should not increase the number of DB queries
        """
        ml_get_mock.return_value.type = self.src.type
        ml_get_mock.return_value.slug = self.src.slug
        ml_get_mock.return_value.name = self.src.name
        ml_get_mock.return_value.version = self.src.revision
        self.src.internal = True
        self.src.save()

        transcriptions = [(
            [[2 * i, 2 * i], [2 * i, 2 * i + 20], [2 * i + 20, 2 * i + 20], [2 * i + 20, 2 * i]],
            str(i),
            i / 100
        ) for i in range(1, 100)]
        data = {
            'element_type': 'text_line',
            'transcription_type': 'line',
            'source': self.src.slug,
            'transcriptions': [{
                'polygon': poly,
                'text': text,
                'score': score
            } for poly, text, score in transcriptions]
        }

        self.client.force_login(self.internal_user)
        with self.assertNumQueries(20):
            response = self.client.post(
                reverse('api:element-transcriptions-bulk', kwargs={'pk': self.page.id}),
                format='json',
                data=data
            )
        self.assertEqual(response.status_code, status.HTTP_201_CREATED)

        created_elts = Element.objects.get_descending(self.page.id)\
            .filter(paths__path__last=self.page.id, type__slug='text_line')
        self.assertEqual(created_elts.count(), 100)
        # Each element has a transcription
        self.assertTrue(all(map(lambda elt: elt.transcriptions.count() == 1, created_elts)))

    @patch('arkindex.project.serializer_fields.MLTool.get')
    @patch('arkindex.project.triggers.get_channel_layer')
    def test_bulk_transcriptions_v2_similar_zone(self, get_layer_mock, ml_get_mock):
        """
        Pushing a transcription matching an element type and zone reuses this element
        """
        get_layer_mock.return_value.send = AsyncMock()
        ml_get_mock.return_value.type = self.src.type
        ml_get_mock.return_value.slug = self.src.slug
        ml_get_mock.return_value.name = self.src.name
        ml_get_mock.return_value.version = self.src.revision
        self.src.internal = True
        self.src.save()

        transcriptions = [
            ([[13, 37], [133, 37], [133, 137], [13, 137], [13, 37]], 'Hello world !', 0.1337),
            (self.line.zone.polygon, 'I <3 JavaScript', 0.42),
        ]
        data = {
            'element_type': 'text_line',
            'transcription_type': 'line',
            'source': self.src.slug,
            'transcriptions': [{
                'polygon': poly,
                'text': text,
                'score': score
            } for poly, text, score in transcriptions]
        }

        self.client.force_login(self.internal_user)
        with self.assertNumQueries(20):
            response = self.client.post(
                reverse('api:element-transcriptions-bulk', kwargs={'pk': self.page.id}),
                format='json',
                data=data
            )
        self.assertEqual(response.status_code, status.HTTP_201_CREATED)

        created_elts = Element.objects.get_descending(self.page.id)
        # The existing text line has been reused
        self.assertEqual(created_elts.count(), 2)
        self.assertCountEqual(self.line.transcriptions.values_list('type', 'text', 'zone'), [
            (TranscriptionType.Line, 'A manual transcription', None),
            (TranscriptionType.Line, 'I <3 JavaScript', None)
        ])

    def test_bulk_transcriptions_v2_requires_verified(self):
        """
        The user has to be verified tu push transcriptions in bulk
        """
        self.user.verified_email = False
        self.user.save()
        self.client.force_login(self.user)
        response = self.client.post(
            reverse('api:element-transcriptions-bulk', kwargs={'pk': self.page.id}),
            format='json',
            data={}
        )
        self.assertEqual(response.status_code, status.HTTP_403_FORBIDDEN)
        self.assertDictEqual(response.json(), {'detail': 'You do not have permission to perform this action.'})

    def test_bulk_transcriptions_v2_private(self):
        """
        Accessing a non writeable element triggers a 404_NOT_FOUND
        """
        self.client.force_login(self.user)
        response = self.client.post(
            reverse('api:element-transcriptions-bulk', kwargs={'pk': self.private_page.id}),
            format='json',
            data={}
        )
        self.assertEqual(response.status_code, status.HTTP_404_NOT_FOUND)

    def test_bulk_transcriptions_v2_non_manual(self):
        """
        Transcriptions source cannot be set to manual
        """
        self.client.force_login(self.user)
        response = self.client.post(
            reverse('api:element-transcriptions-bulk', kwargs={'pk': self.page.id}),
            format='json',
            data={
                'element_type': 'text_line',
                'transcription_type': 'line',
                'source': 'manual',
                'transcriptions': [{
                    'polygon': [[13, 37], [133, 37], [133, 137], [13, 137]],
                    'text': 'Can I write womething here ?',
                    'score': .666
                }]
            }
        )
        self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)
        self.assertDictEqual(response.json(), {
            'source': {'source': ['Transcriptions source slug cannot be set to manual.']}
        })

    def test_bulk_transcriptions_v2_wrong_fields_validation(self):
        """
        Briefly assert fields values are validated
        """
        self.client.force_login(self.user)
        response = self.client.post(
            reverse('api:element-transcriptions-bulk', kwargs={'pk': self.page.id}),
            format='json',
            data={
                'element_type': 'paragraph',
                'transcription_type': 'boulga',
                'source': 'gloubiboulga',
                'transcriptions': [{
                    'polygon': [[13, 37]],
                    'text': 'There is a snake in my computer'
                }]
            }
        )
        self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)
        self.assertDictEqual(response.json(), {
            'element_type': ['Object with slug=paragraph does not exist.'],
            'source': ["Source with slug 'gloubiboulga' not found"],
            'transcription_type': ['Value is not of type TranscriptionType'],
            'transcriptions': [{
                'polygon': [''],
                'score': ['This field is required.']
            }]
        })
