from collections import namedtuple, defaultdict
from pathlib import Path
from django.db.models import CharField, Value, Prefetch
from django.db.models.functions import Concat
from arkindex_common.enums import MetaType
from arkindex.project.polygon import Polygon
from arkindex.project.triggers import reindex_start
from arkindex.images.models import Image
from arkindex.documents.models import Element, ElementType, Corpus, MetaData
import csv
import logging

logger = logging.getLogger(__name__)


ActRow = namedtuple('ActRow', 'number, image_id, coordinates')


class ActsImporter(object):
    """
    Imports acts and surfaces from CSV files in the format (act number, IIIF image ID, surface coordinates)
    Acts without surfaces can be imported with (act number, '', '').
    """

    def __init__(self, csvpath, corpus, folder_type, act_type, surface_type):
        self.csvpath = Path(csvpath)
        assert self.csvpath.is_file()
        assert isinstance(corpus, Corpus)
        assert isinstance(folder_type, ElementType) and folder_type.corpus == corpus and folder_type.folder
        assert isinstance(act_type, ElementType) and act_type.corpus == corpus
        assert isinstance(surface_type, ElementType) and surface_type.corpus == corpus
        self.corpus, self.folder_type, self.act_type, self.surface_type = corpus, folder_type, act_type, surface_type
        self.created_acts_count, self.created_surfaces_count = 0, 0
        self.surface_numbers = defaultdict(int)

    def load(self):
        """
        Prepare some values to perform SQL queries only once
        """
        logger.info('Fetching existing data')

        try:
            self.folder = self.corpus.elements.get(type=self.folder_type, name__iexact=self.csvpath.stem)
            logger.debug('Matched folder "{}" for folder name "{}"'.format(self.folder.name, self.csvpath.stem))
        except Element.DoesNotExist:
            raise ValueError('Volume "{}" not found'.format(self.csvpath.stem))

        # Allows finding an image in the corpus by its IIIF ID,
        # by concatenating the server's base URL and the image path
        self.images = {
            img.image_id: img
            for img in Image.objects
                            .annotate(image_id=Concat('server__url', Value('/'), 'path', output_field=CharField()))
                            .filter(zones__elements__in=Element.objects.get_descending(self.folder.id))
                            .distinct()
        }
        logger.debug('Loaded {} images'.format(len(self.images)))

        # Cannot use get_or_create on acts as we need to look for acts
        # inside the folder itself; we use this dict instead
        self.acts = {}
        acts_queryset = Element.objects \
            .get_descending(self.folder.id) \
            .filter(type=self.act_type, metadatas__name='number') \
            .prefetch_related(Prefetch('metadatas', queryset=MetaData.objects.filter(name='number')))
        for act in acts_queryset:
            # Do not use a QuerySet filter as this would make another query for an already prefetched list
            number = next(md.value for md in act.metadatas.all() if md.name == 'number')
            self.acts[number] = act
        logger.debug('Loaded {} existing acts'.format(len(self.acts)))

        # Same goes for surfaces
        self.surfaces = {
            (surface.image_id, surface.zone.polygon): surface
            for surface in Element.objects
                                  .get_descending(self.folder.id)
                                  .filter(type=self.surface_type, zone__isnull=False)
                                  .prefetch_related('zone')
                                  .annotate(image_id=Concat(
                                      'zone__image__server__url',
                                      Value('/'),
                                      'zone__image__path',
                                      output_field=CharField(),
                                  ))
        }
        logger.debug('Loaded {} existing surfaces'.format(len(self.surfaces)))

    def handle_line(self, row):
        row = ActRow(*row)
        if bool(row.image_id) ^ bool(row.coordinates):
            raise ValueError('Improperly formatted line: either set both image ID and coordinates, or none.')

        act = self.acts.get(row.number)
        if not act:
            act = Element.objects.create(
                type=self.act_type,
                corpus=self.corpus,
                name='Act {}'.format(row.number),
            )
            act.metadatas.create(name='number', type=MetaType.Text, value=row.number)
            act.add_parent(self.folder)
            self.acts[row.number] = act
            self.created_acts_count += 1
            logger.debug('Created act {}'.format(act.name))

        if not row.image_id and not row.coordinates:
            return

        polygon = Polygon.from_str(row.coordinates)
        surface = self.surfaces.get((row.image_id, polygon))
        if not surface:
            image = self.images[row.image_id]
            zone, _ = image.zones.get_or_create(polygon=polygon)

            self.surface_numbers[row.number] += 1

            surface = zone.elements.create(
                type=self.surface_type,
                corpus=self.corpus,
                name='Surface {} on act {}'.format(self.surface_numbers[row.number], row.number),
            )
            self.surfaces[row.image_id, polygon] = surface
            logger.debug('Created surface {}'.format(surface.name))
            self.created_surfaces_count += 1

        surface.add_parent(act)

    def run(self):
        # TODO: Git diff handling
        self.load()
        with self.csvpath.open(newline='') as f:
            reader = csv.reader(f)
            next(reader)  # Skip header line
            rows = [
                ActRow(*row)
                for row in reader
                if row and len(row) == 3
            ]
            count, failed = len(rows), 0

            logger.info('Parsing {} lines'.format(count))
            for row in rows:
                try:
                    self.handle_line(row)
                except Exception as e:
                    logger.warning(str(e))
                    failed += 1

            logger.info('Parsed {} lines and created {} surfaces and {} acts'.format(
                count - failed,
                self.created_surfaces_count,
                self.created_acts_count,
            ))
            assert failed < count, 'No acts were imported'

        logger.info('Updating search index')
        reindex_start(element=self.folder, elements=True)
